#include "../config.hpp"
#include "../game_config.hpp"
#include "../network.hpp"

#include "SDL.h"

#include "game.hpp"
#include "player.hpp"

#include <algorithm>
#include <cassert>
#include <cstdlib>
#include <iostream>
#include <map>
#include <sstream>
#include <vector>

config construct_error(const std::string& msg)
{
	config cfg;
	cfg.add_child("error")["message"] = msg;
	return cfg;
}

void run_server()
{
	const network::manager net_manager;
	const network::server_manager server;

	config login_response;
	login_response.add_child("mustlogin");
	login_response["version"] = game_config::version;

	config initial_response;
	config& gamelist = initial_response.add_child("gamelist");

	typedef std::map<network::connection,player> player_map;
	player_map players;

	game not_logged_in;
	game lobby_players;
	std::vector<game> games;

	for(;;) {
		try {
			network::connection sock = network::accept_connection();
			if(sock) {
				network::send_data(login_response,sock);
				not_logged_in.add_player(sock);
			}

			config data;
			while((sock = network::receive_data(data)) != NULL) {

				std::cerr << "from " << (int)sock << " received: "
				          << data.write() << "\n-----\n";

				//if someone who is not yet logged in is sending
				//login details
				if(not_logged_in.is_member(sock)) {
					const config* const login = data.child("login");

					//client must send a login first.
					if(login == NULL) {
						network::send_data(construct_error(
						                   "You must login first"),sock);
						continue;
					}

					//check the username is valid (all alpha-numeric)
					const std::string& username = (*login)["username"];
					if(std::count_if(username.begin(),username.end(),isalnum)
					   != username.size() || username.empty()) {
						network::send_data(construct_error(
						                   "This username is not valid"),sock);
						continue;
					}

					//check the username isn't already taken
					player_map::const_iterator p;
					for(p = players.begin(); p != players.end(); ++p) {
						if(p->second.name() == username) {
							break;
						}
					}

					if(p != players.end()) {
						network::send_data(construct_error(
						              "This username is already taken"),sock);
						continue;
					}
					
					config* const player_cfg = &initial_response.add_child("user");
					const player new_player(username,*player_cfg);
					players.insert(std::pair<network::connection,player>(sock,new_player));

					//remove player from the not-logged-in list and place
					//the player in the lobby
					not_logged_in.remove_player(sock);
					lobby_players.add_player(sock);

					//currently update user list to all players who are in lobby
					//(later may optimize to only send new login information
					//to all but the new player)
					lobby_players.send_data(initial_response);

				} else if(lobby_players.is_member(sock)) {
					const config* const create_game = data.child("create_game");
					if(create_game != NULL) {

						std::cerr << "creating game...\n";

						//create the new game, remove the player from the
						//lobby and put him/her in the game they have created
						games.push_back(game());
						lobby_players.remove_player(sock);
						games.back().add_player(sock);

						//store the game data here at the moment
						games.back().level() = *create_game;
						std::stringstream converter;
						converter << games.back().id();
						games.back().level()["id"] = converter.str();

						continue;
					}

					//see if the player is joining a game
					const config* const join = data.child("join");
					if(join != NULL) {
						const std::string& id = (*join)["id"];
						const int nid = atoi(id.c_str());
						const std::vector<game>::iterator it =
						             std::find_if(games.begin(),games.end(),
						                          game_id_matches(nid));
						if(it == games.end()) {
							std::cerr << "attempt to join unknown game\n";
							continue;
						}

						lobby_players.remove_player(sock);

						//send them the game data
						network::send_data(it->level(),sock);

						it->add_player(sock);
					}

					//see if it's a message, in which case we add the name
					//of the sender, and forward it to all players in the lobby
					config* const message = data.child("message");
					if(message != NULL) {
						const player_map::const_iterator p = players.find(sock);
						assert(p != players.end());
						(*message)["sender"] = p->second.name();
						lobby_players.send_data(data);
					}

				} else {
					std::vector<game>::iterator g;
					for(g = games.begin(); g != games.end(); ++g) {
						if(g->is_member(sock))
							break;
					}

					if(g == games.end()) {
						std::cerr << "ERROR: unknown socket " << games.size() << "\n";
						continue;
					}

					std::cerr << "data to game: " << data.write() << "\n";

					//if this is data describing the level for a game
					if(data.child("side") != NULL) {

						const bool is_init = g->level_init();

						//if this game is having its level-data initialized
						//for the first time, and is ready for players to join
						if(!is_init) {
							
							//update our config object which describes the
							//open games, and notifies the game of where its description
							//is located at
							config& desc = gamelist.add_child("game",g->level());
							g->set_description(&desc);

							//send all players in the lobby the list of games
							lobby_players.send_data(initial_response);
						}

						//record the new level data, and send to all players
						//who are in the game
						g->level() = data;

						g->send_data(data,sock);
						continue;
					}

					const string_map::const_iterator side =
					                           data.values.find("side");
					if(side != data.values.end()) {
						const bool res = g->take_side(sock,data);
						config response;
						if(res) {
							std::cerr << "played joined side\n";
							response["side_secured"] = side->second;
						} else {
							response["failed"] = "yes";
						}

						network::send_data(response,sock);
						continue;
					}

					if(data.child("start_game")) {
						g->start_game();
					} else if(data.child("leave_game")) {
						const bool needed = g->is_needed(sock);
						g->remove_player(sock);
						lobby_players.add_player(sock);

						if(needed) {

							//tell all other players the game is over,
							//because a needed player has left
							config cfg;
							cfg.add_child("leave_game");
							g->send_data(cfg);

							//delete the game's description
							config* const gamelist = initial_response.child("gamelist");
							assert(gamelist != NULL);
							const config::child_itors vg = gamelist->child_range("game");

							const config::child_iterator desc = std::find(vg.first,vg.second,g->description());
							if(desc != vg.second) {
								delete *desc;
								gamelist->remove_child("game",desc - vg.first);
							}

							//put the players back in the lobby and send
							//them the game list and user list again
							g->send_data(initial_response);
							lobby_players.add_players(*g);
							games.erase(g);
						}

						//send the player who has quit the new game list
						network::send_data(initial_response,sock);

						continue;
					} else if(data["side_secured"].empty() == false) {
						continue;
					} else if(data["failed"].empty() == false) {
						std::cerr << "ERROR: failure to get side\n";
						continue;
					}

					//forward data to all players who are in the game,
					//except for the original data sender
					g->send_data(data,sock);
					g->record_data(data);
				}
			}
			
		} catch(network::error& e) {
			if(!e.socket) {
				std::cerr << "fatal network error: " << e.message << "\n";
				break;
			} else {
				std::cerr << "socket closed: " << e.message << "\n";

				const std::map<network::connection,player>::iterator pl_it =
				               players.find(e.socket);
				if(pl_it != players.end()) {
					const config::child_list& users = initial_response.get_children("user");
					const size_t index = std::find(users.begin(),users.end(),pl_it->second.config_address()) - users.begin();
					if(index < users.size())
						initial_response.remove_child("user",index);

					players.erase(pl_it);
				}
				
				not_logged_in.remove_player(e.socket);
				lobby_players.remove_player(e.socket);
				for(std::vector<game>::iterator i = games.begin();
				    i != games.end(); ++i) {
					if(i->is_needed(e.socket)) {

						//delete the game's configuration
						config* const gamelist =
						            initial_response.child("gamelist");
						assert(gamelist != NULL);
						const config::child_itors vg = gamelist->child_range("game");
						const config::child_list::iterator g = std::find(vg.first,vg.second,i->description());
						if(g != vg.second) {
							const size_t index = g - vg.first;
							delete *g;
							gamelist->remove_child("game",index);
						}
						
						i->disconnect();
						games.erase(i);
						e.socket = 0;
						break;
					} else {
						i->remove_player(e.socket);
					}
				}

				if(e.socket) {
					e.disconnect();
				}

				//send all players the information that a player has logged
				//out of the system
				lobby_players.send_data(initial_response);
			}

			continue;
		} catch(config::error& e) {
			std::cerr << "error in received data: " << e.message << "\n";
			continue;
		}

		SDL_Delay(20);
	}
}

int main()
{
	try {
		run_server();
	} catch(network::error& e) {
		std::cerr << "error starting server: " << e.message << "\n";
		return -1;
	}
}
